# -*- coding: utf-8 -*-
"""Modulo Adaptador Sql basado en Alchemy"""

from abc import ABC
from sqlalchemy import create_engine, text
from sqlalchemy.orm import sessionmaker, scoped_session


class SqlAlchemyAdapter(ABC):
    """Class Adaptador Sql basado en Alchemy"""
    session = None
    entity = None

    def __init__(self, config):
        self._options = config.get_key('db')
        self._session_maker()

    def create(self, data):
        """ Create query """
        try:
            self.session.add(data)
            self.commit()
            return data.id
        except Exception as exception:
            self.rollback()
            raise exception

    def update(self, primay_key, **kwargs):
        """ Create query """
        try:
            self.session.query(self.entity).filter_by(id=primay_key).update(kwargs)
            self.commit()
        except Exception as exception:
            self.rollback()
            raise exception

    def delete(self, primay_key):
        """ Delete query """
        try:
            self.session.query(self.entity).filter_by(
                id=primay_key).delete(synchronize_session=False)
            self.commit()
        except Exception as exception:
            self.rollback()
            raise exception

    def delete_in(self, ids):
        """ Delete por id """
        try:
            self.session.query(self.entity).filter(
                self.entity.id.in_(ids)).delete(synchronize_session=False)
            self.commit()
        except Exception as exception:
            self.rollback()
            raise exception

    def find_all(self, params):
        """ Recuperar toda la lista """
        select_params = params['fields']
        where_params = params['filter']
        pagination_params = params['pagination']
        sort_params = params['sort']

        if select_params:
            columns = [getattr(self.entity, column) for column in select_params]
            query = self.session.query(*columns)
        else:
            query = self.session.query(self.entity)

        if 'filter_in' in params:
            for item_property, values in params['filter_in'].items():
                query = query.filter(getattr(self.entity, item_property).in_(tuple(values)))

        query = query.filter_by(**where_params)
        query = query.order_by(sort_params)
        if pagination_params:
            query = query.offset(pagination_params['offset']).limit(pagination_params['limit'])
        return query.all()

    def find_by_id(self, primary_key):
        """ Recuperar a partir de primary key """
        try:
            return self.session.query(self.entity).filter_by(id=primary_key).first()
        except Exception as exception:
            raise exception

    def statement(self, sql):
        """ Crear query customizada """
        result = self.session.query(self.entity).from_statement(text(sql))
        return result.all()

    def bulk_insert(self, items):
        """ Crear query para insertar """
        try:
            self.session.add_all(items)
            self.commit()
        except Exception as exception:
            self.rollback()
            raise exception

    def delete_with_params(self, params):
        """ Eliminar aplicando where """
        return self.session.query(self.entity).filter(params).delete(synchronize_session=False)

    def commit(self):
        """ Crear un commit transaccional """
        self.session.commit()

    def rollback(self):
        """ Aplicar rollback """
        self.session.rollback()

    def _session_maker(self):
        try:
            driver = 'mysql+pymysql://%s:%s@%s:%s/%s?charset=utf8' % (self._options['user'],
                                                                      self._options['password'],
                                                                      self._options['host'],
                                                                      self._options['port'],
                                                                      self._options['database'])
            engine = create_engine(driver, echo=True, isolation_level="READ UNCOMMITTED")
            self.session = scoped_session(sessionmaker(bind=engine))
        except Exception as exception:
            raise exception



# -*- coding: utf-8 -*-
"""Modulo response"""

class Response(object):
    """Clase response generico"""

    PROCESS_SUCCESS_CODE = 2000
    PROCESS_SUCCESS_MESSAGE = 'SUCCESS'
    PROCESS_ERROR_CODE = 4000
    PROCESS_ERROR_MESSAGE = 'ERROR'

    def success(self, data):
        """Respuesta afirmativa"""
        try:
            return self._schema(False,
                                self.PROCESS_SUCCESS_CODE,
                                self.PROCESS_SUCCESS_MESSAGE,
                                data if data is not None else [])
        except Exception as exception:
            raise exception

    def error(self, message):
        """Respuesta con error"""
        try:
            return self._schema(True,
                                self.PROCESS_ERROR_CODE,
                                self.PROCESS_ERROR_MESSAGE if message is None else message,
                                [])
        except Exception as exception:
            raise exception

    @staticmethod
    def _schema(error, code, message, data):
        return {
            'error': error,
            'code': code,
            'message': message,
            'data': data
        }



# -*- coding: utf-8 -*-
"""Modulo helpers.url"""

def split_query_string(items):
    """dividir un query string"""
    params = {'filter': {}, 'fields': [], 'pagination': {}, 'sort': {}, 'filter_in': {}}
    for key, value in items:
        sub_keys = key.split('.')
        if 'fields' in sub_keys:
            params['fields'] = [value] if isinstance(value, str) else value
        elif sub_keys[0] in params:
            params[sub_keys[0]].update({sub_keys[1]: value})

    params['filter_in'] = get_filter_in(params['filter'])
    params['sort'] = \
        ', '.join("{0} {1}".format(key, val.upper()) for (key, val) in params['sort'].items())

    if params['pagination']:
        params['pagination']['page'] = \
            int(params['pagination']['page']) if 'page' in params['pagination'] else 1
        params['pagination']['limit'] = \
            int(params['pagination']['limit']) if 'limit' in params['pagination'] else 100
        params['pagination'].update({
            'offset':
                (params['pagination']['page'] * \
                params['pagination']['limit']) - \
                params['pagination']['limit']
            })
    return params


def join_query_string():
    """Unir parametros en query string"""
    pass


def get_filter_in(filters):
    """Recuperar filtros de una lista"""
    filter_in = {}
    for filter_index, value in list(filters.items()):
        if isinstance(value, list):
            filter_in.update({filter_index: value})
            del filters[filter_index]
    return filter_in



# -*- coding: utf-8 -*-
""" Helpers Para manejo de excepciones"""

from functools import wraps
from und_microservice.logger.logging import ConsoleLogger

def handler_except(method):
    """Helper para excepcion de los handlers"""
    @wraps(method)
    def method_wrapper(*args, **kwargs):
        """Metodo para los wrappers"""
        try:
            return method(*args, **kwargs)
        except AttributeError as attribute_error:
            raise attribute_error
        except (Exception, ValueError) as exception:
            logger = ConsoleLogger()
            logger.output.error('=== Handler exception ===')
            logger.output.error(exception)
            logger.output.error('=' * 25)
            return exception
    return method_wrapper



# -*- coding: utf-8 -*-
""" Comportamiento base para administrador de configuraciones"""

import glob
from und_microservice.config.base import BaseConfig
from und_microservice.commons.file import File
from und_microservice.helper.data import dict_merge


class ConfigYaml(BaseConfig):
    """ Manejador de config en yml """

    def __init__(self, path='config/'):
        self._config = {}
        yaml_file = File()
        paths = glob.glob(path + '*yml')
        for path in paths:
            config = yaml_file.read_yml(path)
            dict_merge(self._config, config)



# -*- coding: utf-8 -*-
"""Helpers para manipulacion de data"""

import collections
import decimal

def get_object_dict(data_object, params=None):
    """Convertir un objeto en diccionarios"""
    if params is not None and params['fields']:
        fields = params['fields']
    else:
        fields = [attr for attr in data_object.__dict__.keys() if not attr.startswith('_')]
    result = {}
    for field in fields:
        value = getattr(data_object, field)
        if field.startswith('date') and value is not None:
            result.update({field: value.strftime('%Y-%m-%d %H:%M:%S')})
        else:
            result.update({field: parse_value(value)})
    return result


def parse_value(value):
    """Convertir un valor a decimal"""
    return float(value) if isinstance(value, decimal.Decimal) else value


def get_multi_objects_dict(*args, params=None):
    """Convertir un array de objetos en diccionarios"""
    object_group = []
    result = {}
    for data_object in args:
        if params is not None and params['fields']:
            fields = params['fields']
        else:
            fields = [attr for attr in data_object.__dict__.keys() if not attr.startswith('_')]
        row = {}
        for field in fields:
            value = getattr(data_object, field)
            if field.startswith('date') and value is not None:
                row.update({field: value.strftime('%Y-%m-%d %H:%M:%S')})
            else:
                row.update({field: parse_value(value)})
        object_group.append(row)

    for data_object in object_group:
        result.update(**data_object)

    return result

def dict_merge(dct, merge_dct):
    """Mescla diccionarios"""
    for index, item in merge_dct.items():
        if (index in dct and isinstance(dct[index], dict)
                and isinstance(item, collections.Mapping)):
            dict_merge(dct[index], item)
        else:
            dct[index] = item



# -*- coding: utf-8 -*-
"""Logger por consola"""

import logging


class ConsoleLogger(object):
    """Clase logger para trabajar por consola"""

    def __init__(self):
        try:
            self.output = logging
            self.output.basicConfig(
                format='%(asctime)s [%(levelname)s] - %(name)s - %(message)s',
                datefmt='[%Y/%m/%d %I:%M:%S %p]'
            )
        except Exception as exception:
            raise exception



# -*- coding: utf-8 -*-
""" Init Bootstrap """

import codecs
import json
import yaml


class File:
    """ Clase para leer archivos """

    @staticmethod
    def read(file):
        """ Leer texto plano """
        return codecs.open(file, 'r', 'utf-8')

    @staticmethod
    def read_yml(file):
        """ Leer archivos yml """
        file_object = open(file, 'r', encoding='utf-8')
        return yaml.load(file_object)

    @staticmethod
    def read_json(file):
        """ Leer archivos json """
        file_object = open(file, 'r')
        return json.load(file_object)



# -*- coding: utf-8 -*-
"""Helpers para manejo de strings"""

def concat(item_list, separator="-"):
    """Concatena un array de string con un separador"""
    return separator.join(str(item) for item in item_list)



# -*- coding: utf-8 -*-
""" Comportamiento base para administrador de configuraciones"""

from abc import ABC
from und_microservice.config.exception.config import NotConfigException
from und_microservice.config.exception.config import NotValueConfigException

class  BaseConfig(ABC):
    """ Clase de configuracion base """
    _config = None

    def get_all(self):
        """ Permite recuperar la todas las configuraciones """
        if self._config == '':
            raise NotConfigException
        return self._config

    def get_key(self, key):
        """ Permite recuperar un grupo de configuraciones """
        if self._config == '':
            raise NotValueConfigException
        return self._config[key]
